# from: https://gitlab.com/abraxos/click-path/-/blob/a1536f8a2dcbcee026e77a23cc515fe0decb824d/click_path/core.py
from glob import iglob
from os import R_OK, W_OK, access
from pathlib import Path
from typing import Iterable, Tuple

import rich_click as click
from typeguard import typechecked


class GlobPaths(click.ParamType):
    """
    This class is designed to allow users of a click application to pass in a
    glob-based file pattern with filtering and other various requirements.
    While some of the parameters are very similar to those found in the
    click.Path type, they are more geared towards filtering out files that do
    not fit the given requirements and non-existent files are not really
    meaningful for globbing.
    """

    name = "Path-or-Glob"

    @typechecked
    def __init__(
        self,
        files_okay: bool = True,
        dirs_okay: bool = True,
        writable_only: bool = False,
        readable_only: bool = True,
        resolve: bool = False,
        at_least_one: bool = True,
    ):
        self.files_okay = files_okay
        self.dirs_okay = dirs_okay
        self.writable_only = writable_only
        self.readable_only = readable_only
        self.resolve = resolve
        self.at_least_one = at_least_one

    @typechecked
    def _validated_path(self, file_path: Path) -> Tuple[Path, bool, str]:
        if not file_path.exists():
            return file_path, False, "does not exist"
        if self.writable_only and not access(str(file_path), W_OK):
            return file_path, False, "exists, but is not writable"
        if self.readable_only and not access(str(file_path), R_OK):
            return file_path, False, "exists, but is not readable"
        if not self.files_okay and file_path.is_file():
            return file_path, False, "exists, is a file"
        if not self.dirs_okay and file_path.is_dir():
            return file_path, False, "exists, is a directory"
        return file_path, True, "valid"

    @typechecked
    def convert(
        self, value: str, param: click.Parameter, ctx: click.Context
    ) -> Iterable[Path]:
        validation_results = [self._validated_path(Path(p)) for p in iglob(value)]
        if self.at_least_one and not any(valid for _, valid, __ in validation_results):
            summary = "\n".join(
                f"\t{path}: {why}" for path, _, why in validation_results
            )
            self.fail(
                f'No paths from "{value}" matched the requirements: [{", ".join(self.filters())}]\n{summary}',
                param,
                ctx,
            )
        return [
            path if not self.resolve else path.resolve()
            for path, _, __ in validation_results
        ]

    @typechecked
    def filters(self) -> Iterable[str]:
        filters = []
        if self.files_okay:
            filters.append("files")
        if self.dirs_okay:
            filters.append("directories")
        if self.writable_only:
            filters.append("write-permissions")
        if self.readable_only:
            filters.append("read-permissions")
        return filters



import binascii
from typing import Final

import imagehash
import numpy as np
from imagehash import ImageHash
from PIL import Image

FRAME_SIZE: Final[int] = 32


def imghash_to_bytes(imghash: ImageHash) -> bytes:
    """
    >>> import numpy as np
    >>> imghash_to_bytes(ImageHash(np.array([\
        np.array([ True,  True, False,  True, False,  True, False,  True]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False])])))
    b'\\xd5\\x00\\x00\\x00\\x00\\x00\\x00\\x00'
    """
    return binascii.a2b_hex(str(imghash))


def rawframe_to_imghash(
    raw_frame: bytes,
    frame_width: int = FRAME_SIZE,
    frame_height: int = FRAME_SIZE,
) -> imagehash.ImageHash:
    """Apply a Perceptual Hash computation on raw image frame.

    >>> rawframe_to_imghash(bytearray((128,)*(FRAME_SIZE**2)))
    array([[ True, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False]])
    """
    return imagehash.phash(
        Image.fromarray(
            # https://docs.scipy.org/doc/numpy-1.14.0/reference/generated/numpy.frombuffer.html
            np.frombuffer(
                raw_frame, dtype=np.uint8, count=frame_width * frame_height
            ).reshape(frame_width, frame_height)
        )
    )



import datetime
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterator, Tuple

from imageio_ffmpeg import count_frames_and_secs, read_frames

from vhcalc.tools.imghash import FRAME_SIZE


@dataclass
class MetaData:
    fps: float
    duration: float
    nb_frames: int


def build_reader_frames(
    media: Path,
    nb_seconds_to_extract: float = 0,
    seek_to_middle: bool = False,
    ffmpeg_reduce_verbosity: bool = True,
) -> Tuple[Iterator[bytes], MetaData]:
    s_media = str(media)
    meta: dict[str, Any] = next(read_frames(s_media))
    ffmpeg_seek_input_cmd: list[str] = []
    ffmpeg_seek_output_cmd: list[str] = []

    # extract a (frame's) chunk around/in middle of the media
    # https://trac.ffmpeg.org/wiki/Seeking#Cuttingsmallsections
    if ffmpeg_reduce_verbosity:
        ffmpeg_seek_input_cmd += "-hide_banner -nostats -nostdin".split(" ")

    if seek_to_middle:
        # it's an approximation, this command seek around/close to the middle
        ffmpeg_seek_input_cmd += (
            "-ss",
            str(datetime.timedelta(seconds=meta["duration"] // 2)),
        )

    if nb_seconds_to_extract:
        # express in number of frames to extract (more precise)
        ffmpeg_seek_output_cmd += (
            "-frames:v",
            str(round(nb_seconds_to_extract * meta["fps"])),
        )
    # rescale output frame to 32x32
    video_filters = f"scale=width={FRAME_SIZE}:height={FRAME_SIZE}"

    reader = read_frames(
        s_media,
        input_params=[*ffmpeg_seek_input_cmd],
        output_params=[
            *ffmpeg_seek_output_cmd,
            *("-vf", video_filters),
            *("-pix_fmt", "gray"),
        ],
        bits_per_pixel=8,
    )
    # yield metadata
    next(reader)

    # build metadata dataclass
    metadata = MetaData(
        fps=meta["fps"],
        duration=meta["duration"],
        nb_frames=count_frames_and_secs(media)[0],
    )

    return reader, metadata



from pathlib import Path

# https://pypi.org/project/click-pathlib/
from tempfile import gettempdir
from typing import Iterable, Optional

from rich import get_console

from vhcalc.services.reader_frames import build_reader_frames
from vhcalc.tools.chunk import chunks
from vhcalc.tools.imghash import imghash_to_bytes, rawframe_to_imghash
from vhcalc.tools.progress_bar import configure_progress_bar

console = get_console()


def compute_imghash_from_media(
    input_media: Path,
    chunk_nb_seconds: int = 15,
) -> Iterable[bytes]:
    # Read a video file
    it_reader_frame, media_metadata = build_reader_frames(input_media)
    chunk_size = int(media_metadata.fps * chunk_nb_seconds)

    # configure chunk
    gen_imghashes = map(rawframe_to_imghash, it_reader_frame)
    gen_chunk_imghashes = chunks(gen_imghashes, chunk_size)
    # for each chunk of frames
    for chunk_imghashes in gen_chunk_imghashes:
        for frame_hash_binary in map(imghash_to_bytes, chunk_imghashes):
            # and write (chunk of) images hashes result on export file
            yield frame_hash_binary


def export_imghash_from_media(
    input_media: Path,
    output_file: Optional[Path] = None,
    chunk_nb_seconds: int = 15,
    unlink_export_file: bool = True,
) -> Path:
    """

    :param input_media:
    :param output_file:
    :param chunk_nb_seconds:
    :param unlink_export_file:

    :return:
    """
    # Read a video file
    it_reader_frame, media_metadata = build_reader_frames(input_media)
    nb_frames_to_read = media_metadata.nb_frames
    chunk_size = int(media_metadata.fps * chunk_nb_seconds)

    console.print(f"{media_metadata}")
    console.print(f"Number of frames to read: {nb_frames_to_read}")
    console.print(f"Chunk size (nb frames): {chunk_size}")

    # manage export
    if not output_file:
        # https://bandit.readthedocs.io/en/latest/plugins/b108_hardcoded_tmp_directory.html
        output_file = (
            Path(gettempdir()) / f"{input_media.name}.{media_metadata.fps}fps.phash"
        )

    # remove/unlink export file if exist
    if unlink_export_file:
        output_file.unlink(missing_ok=True)

    console.print(f"output_file: {str(output_file)}")

    # configure progress bar
    progress_bar, pb_task_id = configure_progress_bar(
        input_media.name, nb_frames_to_read * 8, console
    )
    pb_advance = chunk_size * 8

    with progress_bar:
        gen_imghashes = map(rawframe_to_imghash, it_reader_frame)
        gen_chunk_imghashes = chunks(gen_imghashes, chunk_size)
        # for each chunk of frames
        for chunk_imghashes in gen_chunk_imghashes:
            # ... open/write and close export file in binary append mode
            with output_file.open("ab") as fo:
                # consume generator: chunk frame to image hash
                for frame_hash_binary in map(imghash_to_bytes, chunk_imghashes):
                    # and write (chunk of) images hashes result on export file
                    fo.write(frame_hash_binary)
            # update progress bar synchronize with chunk progression
            progress_bar.update(pb_task_id, advance=pb_advance, refresh=True)
    return output_file



import sys

import pkg_resources
from imageio_ffmpeg import get_ffmpeg_version


def get_version_extended_informations() -> str:
    return "\n".join(
        (
            f"python version: {sys.version}",
            "\n".join(
                f"{pkg_distribution.project_name}: {pkg_distribution.version}"
                for pkg_distribution in map(
                    pkg_resources.get_distribution, ("imagehash", "numpy", "pillow")
                )
            ),
            f"ffmpeg version {get_ffmpeg_version()}",
        )
    )



# -*- coding: utf-8 -*-
import pathlib
import sys
from importlib.metadata import version
from typing import BinaryIO, Iterable, Optional

import rich_click as click
from loguru import logger

import vhcalc.services as services
from vhcalc.tools.click_default_group import DefaultGroup
from vhcalc.tools.click_path import GlobPaths
from vhcalc.tools.version_extended_informations import get_version_extended_informations


@click.version_option(
    version=version("vhcalc"),
    prog_name="vhcalc",
    message=f"%(prog)s, version %(version)s\n{get_version_extended_informations()}",
)
@click.group(
    cls=DefaultGroup,
    default="imghash",
    default_if_no_args=True,
    invoke_without_command=True,
)
def cli() -> None:
    pass


@cli.command(
    short_help="extracting and exporting binary video hashes (fingerprints) from any video source"
)
@click.argument("media_filename", type=click.Path(exists=True))
@click.argument("output_stream", type=click.File("wb"), default=sys.stdout.buffer)
def imghash(media_filename: str, output_stream: BinaryIO) -> None:
    """Simple form of the application: Input filepath > image hashes (to stdout by default)"""
    for frame_hash_binary in services.compute_imghash_from_media(
        pathlib.Path(media_filename)
    ):
        output_stream.write(frame_hash_binary)


@cli.command(
    short_help="extracting and exporting binary video hashes (fingerprints) from any video source"
)
@click.option(
    "--medias_pattern",
    "-r",
    required=True,
    type=GlobPaths(
        files_okay=True,
        dirs_okay=False,
        readable_only=True,
        at_least_one=True,
    ),
    help="Pattern to find medias",
)
@click.option(
    "--output-file",
    "-o",
    default=None,
    type=click.Path(writable=True, path_type=pathlib.Path),
    help="File where to write images hashes.",
)
@logger.catch
def export_imghash_from_media(
    medias_pattern: Iterable[pathlib.Path], output_file: Optional[pathlib.Path]
) -> None:
    """Click entrypoint for extracting and exporting binary video hashes (fingerprints) from any video source"""
    for media in medias_pattern:
        services.export_imghash_from_media(media, output_file)


if __name__ == "__main__":
    cli()



"""
   click_default_group
   ~~~~~~~~~~~~~~~~~~~

   Define a default subcommand by `default=True`:

   .. sourcecode:: python

      import click
      from click_default_group import DefaultGroup

      @click.group(cls=DefaultGroup, default_if_no_args=True)
      def cli():
          pass

      @cli.command(default=True)
      def foo():
          click.echo('foo')

      @cli.command()
      def bar():
          click.echo('bar')

   Then you can invoke that without explicit subcommand name:

   .. sourcecode:: console

      $ cli.py --help
      Usage: cli.py [OPTIONS] COMMAND [ARGS]...

      Options:
        --help    Show this message and exit.

      Command:
        foo*
        bar

      $ cli.py
      foo
      $ cli.py foo
      foo
      $ cli.py bar
      bar

"""

import typing as t
import warnings

import click

__all__ = ["DefaultGroup"]
__version__ = "1.2.4"


class DefaultGroup(click.Group):
    """Invokes a subcommand marked with `default=True` if any subcommand not
    chosen.

    :param default_if_no_args: resolves to the default command if no arguments
                               passed.

    """

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        # To resolve as the default command.
        if not kwargs.get("ignore_unknown_options", True):
            raise ValueError("Default group accepts unknown options")
        self.ignore_unknown_options = True
        self.default_cmd_name = kwargs.pop("default", None)
        self.default_if_no_args = kwargs.pop("default_if_no_args", False)
        super().__init__(*args, **kwargs)

    def set_default_command(self, command: t.Any) -> None:
        """Sets a command function as the default command."""
        cmd_name = command.name
        self.add_command(command)
        self.default_cmd_name = cmd_name

    def parse_args(self, ctx: click.core.Context, args: t.List[str]) -> t.List[str]:
        if not args and self.default_if_no_args:
            args.insert(0, self.default_cmd_name)
        return super().parse_args(ctx, args)

    def get_command(
        self, ctx: click.core.Context, cmd_name: str
    ) -> t.Optional[click.core.Command]:
        if cmd_name not in self.commands:
            # No command name matched.
            ctx.arg0 = cmd_name  # type: ignore
            cmd_name = self.default_cmd_name
        return super().get_command(ctx, cmd_name)

    def resolve_command(
        self, ctx: click.core.Context, args: t.List[str]
    ) -> t.Tuple[t.Optional[str], t.Optional[click.core.Command], t.List[str]]:
        cmd_name, cmd, args = super().resolve_command(ctx, args)
        if cmd and hasattr(ctx, "arg0"):
            args.insert(0, ctx.arg0)
            cmd_name = cmd.name
        return cmd_name, cmd, args

    def format_commands(
        self, ctx: click.core.Context, formatter: click.formatting.HelpFormatter
    ) -> None:
        new_formatter = DefaultCommandFormatter(self, formatter, mark="*")
        return super().format_commands(ctx, new_formatter)

    @t.overload
    def command(self, __func: t.Callable[..., t.Any]) -> click.core.Command: ...

    @t.overload
    def command(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.Callable[[t.Callable[..., t.Any]], click.core.Command]: ...

    def command(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.Union[
        t.Callable[[t.Callable[..., t.Any]], click.core.Command], click.core.Command
    ]:
        default = kwargs.pop("default", False)
        decorator: t.Callable[[t.Callable[..., t.Any]], click.core.Command] = (
            super().command(*args, **kwargs)
        )
        if not default:
            return decorator
        warnings.warn(
            "Use default param of DefaultGroup or set_default_command() instead",
            DeprecationWarning,
        )

        def _decorator(f: t.Callable[..., t.Any]) -> click.core.Command:
            cmd: click.core.Command = decorator(f)
            self.set_default_command(cmd)
            return cmd

        return _decorator


class DefaultCommandFormatter(click.formatting.HelpFormatter):
    """Wraps a formatter to mark a default command."""

    def __init__(
        self,
        group: DefaultGroup,
        formatter: click.formatting.HelpFormatter,
        mark: str = "*",
    ):
        self.group = group
        self.formatter = formatter
        self.mark = mark

        super().__init__()

    def __getattr__(self, attr: str) -> t.Any:
        return getattr(self.formatter, attr)

    def write_dl(
        self, rows: t.Sequence[t.Tuple[str, str]], *args: t.Any, **kwargs: t.Any
    ) -> None:
        rows_: t.List[t.Tuple[str, str]] = []
        for cmd_name, text in rows:
            if cmd_name == self.group.default_cmd_name:
                rows_.insert(0, (cmd_name + self.mark, text))
            else:
                rows_.append((cmd_name, text))
        return self.formatter.write_dl(rows_, *args, **kwargs)
